"""
This example shows how to run offline inference using continuous batching.

NOTE: At the moment, if you are checking parity, things may not line up
unless you compare eager against the FMS cpu model, i.e.,
    $ python cb_spyre_vision.py --backend eager --compare-target fms
"""

import argparse
import os
import platform
import time

import torch
from fms.models import get_model
from fms.utils import serialization
from fms.utils.generation import generate as fms_generate
from transformers import AutoModelForVision2Seq, AutoProcessor
from vllm import LLM, SamplingParams
from vllm.assets.image import ImageAsset

parser = argparse.ArgumentParser()
parser.add_argument("--model", type=str, default="ibm-granite/granite-vision-3.3-2b")
parser.add_argument(
    "--max_model_len", "--max-model-len", type=int, default=8192
)  # one image has a max context of ~5k
parser.add_argument("--max_num_seqs", "--max-num-seqs", type=int, default=2)
parser.add_argument("--tp", type=int, default=1)
parser.add_argument("--num-prompts", "-n", type=int, default=1)
parser.add_argument(
    "--max-tokens",
    type=str,
    default="8",
    help="Comma separated list of max tokens to use for each prompt. "
    "This list is repeated until prompts are exhausted.",
)
parser.add_argument("--backend", type=str, default="sendnn", choices=["eager", "sendnn"])

parser.add_argument(
    "--compare-target",
    type=str,
    default="fms",
    choices=["transformers", "fms"],
    help="Target to compare results against on CPU.",
)


def get_vllm_prompts(num_prompts):
    """Get the vLLM prompts to be processed."""
    template = "<|system|>\nA chat between a curious user and an artificial intelligence assistant. The assistant gives helpful, detailed, and polite answers to the user's questions.\n<|user|>\n<image>\n{}\n<|assistant|>\n"  # noqa: E501

    images = [
        ImageAsset("cherry_blossom").pil_image,
        ImageAsset("stop_sign").pil_image,
    ]

    instructions = [
        "describe this image.",
        "what is shown in this image?",
        "what kind of flowers are these?",
    ]

    prompts = []
    for img in images:
        width, height = img.size
        for instr in instructions:
            # Make the images smol so that this example can run faster,
            # since we are not using a toy model here, and big images
            # can take up tons of tokens
            new_width = int(0.1 * width)
            new_height = int(0.1 * height)
            prompts.append(
                {
                    "prompt": template.format(instr),
                    "multi_modal_data": {
                        "image": img.resize((new_width, new_height)),
                    },
                }
            )

    prompts = prompts * (num_prompts // len(prompts) + 1)
    return prompts[:num_prompts]


def compare_results(
    prompts: list[str], outputs_a: list[str], outputs_b: list[str], name_a: str, name_b: str
):
    """Utils for comparing outputs from differing engines/implementations,
    e.g., transformers & vLLM.
    """

    print(f"Comparing {name_a} results with {name_b}")
    print("===============")
    any_differ = False
    for idx, (result_a, result_b) in enumerate(zip(outputs_a, outputs_b)):
        if result_a != result_b:
            img_tok_idx = prompts[idx].index("<image>")
            gen_prompt_idx = prompts[idx].index("<|assistant|>")
            raw_prompt = prompts[idx][img_tok_idx:gen_prompt_idx].strip()

            any_differ = True
            print(f"Results for prompt {idx} differ!")
            print(f"\nPrompt (no system/gen prompt):\n {repr(raw_prompt)}")
            print(f"\n{name_a} generated text:\n {result_a}\n")
            print(f"\n{name_b} generated text:\n {result_b}\n")
            print("-----------------------------------")

    if not any_differ:
        print("\nAll results match!\n")


### Alternate implementations to compare against
def get_transformers_results(model_path, vllm_prompts):
    """Process the results for HF Transformers running on CPU."""
    model = AutoModelForVision2Seq.from_pretrained(model_path)
    return process_prompts(
        model_path,
        model,
        vllm_prompts,
        process_prompt_transformers,
    )


def process_prompt_transformers(model, max_tokens, inputs):
    """Process a single prompt using a transformers model."""
    return model.generate(**inputs, max_new_tokens=max_tokens)


def get_fms_results(model_path, vllm_prompts):
    """Process the results for FMS running on CPU."""
    # head_dim expansion required for granite vision
    serialization.extend_adapter("llava_next", "hf", ["weight_expansion_for_mismatched_head_dim"])
    config_dict = {}
    config_dict["head_dim"] = 128

    # Load, but don't compile (compare to CPU)
    model = get_model(
        "hf_pretrained",
        model_path,
        data_type=torch.bfloat16,  # Matches default in vLLM for this model
        fused_weights=False,
        override_hf_pretrained_config=True,
        text_config=config_dict,
    )

    return process_prompts(
        model_path,
        model,
        vllm_prompts,
        process_prompt_fms,
    )


def process_prompt_fms(model, max_tokens, inputs):
    """Process a single prompt using an FMS model."""
    input_ids = inputs.pop("input_ids")
    # May be better to use paged attn later on, but for now
    # we just use sdpa to avoid having to deal with padding
    # utils & position id management here
    inputs["attn_name"] = "sdpa_causal"

    return fms_generate(
        model,
        input_ids,
        max_new_tokens=max_tokens,
        use_cache=True,
        do_sample=False,  # Greedy decode
        extra_kwargs=inputs,
        prepare_model_inputs_hook=model.prepare_inputs_for_generation,
    )


def process_prompts(model_path, model, vllm_prompts, process_prompt):
    """Generic wrapper for running generate on either transformers or FMS."""
    processor = AutoProcessor.from_pretrained(model_path)
    num_prompts = len(vllm_prompts)
    generated_texts = []
    for i in range(num_prompts):
        # Prompts are preformatted, so don't worry about the chat template
        vllm_req = vllm_prompts[i]

        inputs = processor(
            text=vllm_req["prompt"],
            images=vllm_req["multi_modal_data"]["image"],
            return_tensors="pt",
        )
        # NOTE: Image tokens are expanded in the llava next preprocessor
        num_expanded_toks = inputs.input_ids.shape[1]

        target_output = process_prompt(
            model,
            max_tokens[i],
            inputs,
        )

        out_toks = target_output[0][num_expanded_toks:]
        # Make sure not to include EOS, since vLLM
        # doesn't return them, but FMS might.
        generated_text = processor.decode(
            out_toks,
            skip_special_tokens=True,
        )
        generated_texts.append(generated_text)

    return generated_texts


if __name__ == "__main__":
    args = parser.parse_args()

    max_num_seqs = args.max_num_seqs  # defines the max batch size

    if platform.machine() == "arm64":
        print(
            "Detected arm64 running environment. "
            "Setting HF_HUB_OFFLINE=1 otherwise vllm tries to download a "
            "different version of the model using HF API which might not work "
            "locally on arm64."
        )
        os.environ["HF_HUB_OFFLINE"] = "1"

    os.environ["VLLM_SPYRE_DYNAMO_BACKEND"] = args.backend
    os.environ["VLLM_SPYRE_USE_CB"] = "1"
    os.environ["VLLM_SPYRE_USE_CHUNKED_PREFILL"] = "1"

    prompts = get_vllm_prompts(args.num_prompts)

    # Set differing max_tokens so that the requests drop out of the batch at
    # different times
    max_tokens = [int(v) for v in args.max_tokens.split(",")]
    max_tokens = max_tokens * (args.num_prompts // len(max_tokens) + 1)
    max_tokens = max_tokens[: args.num_prompts]

    sampling_params = [
        SamplingParams(max_tokens=m, temperature=0.0, ignore_eos=True) for m in max_tokens
    ]

    llm = LLM(
        model=args.model,
        tokenizer=args.model,
        max_model_len=args.max_model_len,
        max_num_seqs=max_num_seqs,
        tensor_parallel_size=args.tp,
    )

    # Generate texts from the prompts. The output is a list of RequestOutput
    # objects that contain the prompt, generated text, and other information.
    print("=============== GENERATE")
    t0 = time.time()
    vllm_outputs = llm.generate(prompts, sampling_params)
    vllm_results = [x.outputs[0].text for x in vllm_outputs]  # raw texts
    raw_prompts = [prompt["prompt"] for prompt in prompts]

    compare_target_map = {
        "transformers": get_transformers_results,
        "fms": get_fms_results,
    }

    # Since we always compare the results here, we don't bother
    # printing the raw results yet, since the head_dim patch
    # in FMS init tends to flood the logs anyway.
    cpu_results = compare_target_map[args.compare_target](
        model_path=args.model,
        vllm_prompts=prompts,
    )

    compare_results(
        prompts=raw_prompts,
        outputs_a=cpu_results,
        outputs_b=vllm_results,
        name_a=f"{args.compare_target} [cpu]",
        name_b="vllm [spyre]",
    )



import pytest
from pydantic import ValidationError
from vllm.entrypoints.openai.cli_args import make_arg_parser
from vllm import EngineArgs


from vllm_spyre.platform import SpyrePlatform
from spyre_util import environ_checkpoint, REFERENCE_MODELS

try:
    # old
    from vllm.utils import FlexibleArgumentParser
except ImportError:
    # new
    from vllm.utils.argparse_utils import FlexibleArgumentParser

global_default = 192


# The default chunk size we set for all models when chunked prefill
# is enabled is 1024. For some models (granite-3.3-8b-instruct)
# we force a model-specific value that the user can't change except
# by using VLLM_DT_CHUNK_LEN. In the case of granite the default is
# also 1024, so to be able to verify that the special model logic
# is working, we change the default to 192 (3 blocks), which is very
# unlikely to be used for specific models in the future.
# In the parametrization below, micro-g3 represents a generic model
# which will use the global default.
@pytest.mark.parametrize(
    "model_name, chunk_size",
    [
        ("ibm-ai-platform/micro-g3.3-8b-instruct-1b", global_default),
        ("ibm-granite/granite-3.3-8b-instruct", 1024),
    ],
)
def test_generic_model_chunk_size_default(
    model_name: str, chunk_size: int, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setenv("VLLM_SPYRE_USE_CB", "1")

    # Change the default so that we can differentiate the global
    # default from model-specific defaults.
    monkeypatch.setattr(SpyrePlatform, "DEFAULT_CHUNK_SIZE", global_default)

    # Some configuration code paths are only activate with sendnn and tp=4.
    # We want to enable them because they are the ones we care about for production.
    # But we need to patch sendnn_configured to prevent exceptions
    monkeypatch.setenv("VLLM_SPYRE_DYNAMO_BACKEND", "sendnn")

    def sendnn_configured() -> bool:
        return False

    monkeypatch.setattr(SpyrePlatform, "sendnn_configured", sendnn_configured)
    # We always use the micro model because we can't download all model
    # configurations in the GHA environment, but we patch SpyrePlatform
    # to test the behavior for all models.
    model = REFERENCE_MODELS["ibm-ai-platform/micro-g3.3-8b-instruct-1b"]
    common_args = [
        "--model",
        model.name,
        "--revision",
        model.revision,
        "--max-model-len",
        "64",
        "--max-num-seqs",
        "32",
        "-tp",
        "4",
        "--swap-space",  # to prevent a validation error in the 16GB memory test env.
        "1",
    ]

    if model_name == "ibm-granite/granite-3.3-8b-instruct":
        monkeypatch.setattr(SpyrePlatform, "is_granite_3_8b", lambda _config: True)

    with environ_checkpoint():
        # Test that the upstream default is None but is changed to 2048 by
        # the VllmConfig initialization (when using CLI), but is then
        # overridden in our platform.py
        engine_args = _build_engine_args(
            [
                *common_args,
            ]
        )

        assert engine_args.max_num_batched_tokens is None
        vllm_config = engine_args.create_engine_config()
        # TODO: this behavior of changing the engine_args was introduced in v0.12.0.
        # Uncomment below when this version becomes the lowest supported version.
        # assert engine_args.max_num_batched_tokens == 2048
        # we override this in platform.py
        assert (
            vllm_config.scheduler_config.max_num_batched_tokens
            == vllm_config.model_config.max_model_len * vllm_config.scheduler_config.max_num_seqs
        )

    with environ_checkpoint():
        # Test that we override the default and even the user provided setting
        engine_args = _build_engine_args([*common_args, "--max-num-batched-tokens", "128"])
        assert engine_args.max_num_batched_tokens == 128
        vllm_config = engine_args.create_engine_config()
        assert engine_args.max_num_batched_tokens == 128
        # we override this in platform.py
        assert (
            vllm_config.scheduler_config.max_num_batched_tokens
            == vllm_config.model_config.max_model_len * vllm_config.scheduler_config.max_num_seqs
        )

    # Enable CP
    monkeypatch.setenv("VLLM_SPYRE_USE_CHUNKED_PREFILL", "1")

    with environ_checkpoint():
        # Test that the default is 192 when CP is enabled
        engine_args = _build_engine_args(
            [
                *common_args,
            ]
        )
        assert engine_args.max_num_batched_tokens == 192
        vllm_config = engine_args.create_engine_config()
        assert engine_args.max_num_batched_tokens == 192
        assert vllm_config.scheduler_config.max_num_batched_tokens == chunk_size

    with environ_checkpoint():
        # Test that we can still change the default
        engine_args = _build_engine_args([*common_args, "--max-num-batched-tokens", "128"])
        assert engine_args.max_num_batched_tokens == 128
        vllm_config = engine_args.create_engine_config()
        assert engine_args.max_num_batched_tokens == 128
        if chunk_size == global_default:
            assert vllm_config.scheduler_config.max_num_batched_tokens == 128
        else:
            assert vllm_config.scheduler_config.max_num_batched_tokens == chunk_size

    with environ_checkpoint():
        # Test that an invalid value will trigger an error (42 is not a multiple of the block size)
        engine_args = _build_engine_args([*common_args, "--max-num-batched-tokens", "42"])
        assert engine_args.max_num_batched_tokens == 42
        if chunk_size == global_default:
            with pytest.raises(ValidationError):
                vllm_config = engine_args.create_engine_config()
        else:
            assert vllm_config.scheduler_config.max_num_batched_tokens == chunk_size

    monkeypatch.setenv("VLLM_DT_CHUNK_LEN", "512")

    with environ_checkpoint():
        # Verify that VLLM_DT_CHUNK_LEN overrides the default
        engine_args = _build_engine_args(
            [
                *common_args,
            ]
        )
        assert engine_args.max_num_batched_tokens == 192
        vllm_config = engine_args.create_engine_config()
        assert engine_args.max_num_batched_tokens == 192
        assert vllm_config.scheduler_config.max_num_batched_tokens == 512

    with environ_checkpoint():
        # Verify that VLLM_DT_CHUNK_LEN overrides the cli setting
        engine_args = _build_engine_args([*common_args, "--max-num-batched-tokens", "64"])
        assert engine_args.max_num_batched_tokens == 64
        vllm_config = engine_args.create_engine_config()
        assert engine_args.max_num_batched_tokens == 64
        assert vllm_config.scheduler_config.max_num_batched_tokens == 512


def test_prefix_caching_is_off_by_default(monkeypatch: pytest.MonkeyPatch) -> None:
    model = REFERENCE_MODELS["ibm-ai-platform/micro-g3.3-8b-instruct-1b"]
    common_args = [
        "--model",
        model.name,
        "--revision",
        model.revision,
        "--max-model-len",
        "1024",
    ]

    # only care about prefix caching with chunked prefill
    monkeypatch.setenv("VLLM_SPYRE_USE_CB", "1")
    monkeypatch.setenv("VLLM_SPYRE_USE_CHUNKED_PREFILL", "1")

    with (
        environ_checkpoint(),
        pytest.MonkeyPatch.context() as m,
    ):
        # Verify that prefix caching is on by default unless we change that in
        # pre_register_and_update
        def pre_register_and_update(cls, parser: FlexibleArgumentParser | None = None) -> None:
            pass

        m.setattr(SpyrePlatform, "pre_register_and_update", pre_register_and_update)

        engine_args = _build_engine_args(
            [
                *common_args,
            ]
        )
        assert engine_args.enable_prefix_caching is None
        vllm_config = engine_args.create_engine_config()
        assert engine_args.enable_prefix_caching
        assert vllm_config.cache_config.enable_prefix_caching

    with environ_checkpoint():
        # Verify that prefix caching is off by default with vllm-spyre
        engine_args = _build_engine_args(
            [
                *common_args,
            ]
        )
        assert not engine_args.enable_prefix_caching
        vllm_config = engine_args.create_engine_config()
        assert not engine_args.enable_prefix_caching
        assert not vllm_config.cache_config.enable_prefix_caching

    with environ_checkpoint():
        # Test that it can be enabled
        engine_args = _build_engine_args([*common_args, "--enable-prefix-caching"])
        assert engine_args.enable_prefix_caching
        vllm_config = engine_args.create_engine_config()
        assert engine_args.enable_prefix_caching
        assert vllm_config.cache_config.enable_prefix_caching


def _build_engine_args(cli_args: list[str]) -> EngineArgs:
    parser = FlexibleArgumentParser(description="vLLM's remote OpenAI server.")
    parser = make_arg_parser(parser)
    args = parser.parse_args(cli_args)

    engine_args = EngineArgs.from_cli_args(args)

    return engine_args



"""Tests validating the correctness and configuration of prompt_logprobs.

Run `python -m pytest tests/e2e/test_spyre_prompt_logprobs.py`.
"""

import math

import pytest
import torch
import torch.nn.functional
from llm_cache_util import force_engine_shutdown
from spyre_util import ModelInfo, get_chicken_soup_prompts, skip_unsupported_tp_size
from transformers import AutoModelForCausalLM, AutoTokenizer
from vllm import LLM, RequestOutput, SamplingParams
from vllm.config import ModelConfig, VllmConfig

from vllm_spyre.platform import SpyrePlatform


# Skip for now until prompt logprobs are fixed
@pytest.mark.skip
def test_prompt_logprobs(
    backend: str, model: ModelInfo, tp_size: int, monkeypatch: pytest.MonkeyPatch
) -> None:
    """
    This test checks the prompt_logprobs output from vllm against a reference
    implementation using huggingface.
    """
    skip_unsupported_tp_size(tp_size, backend)
    if "FP8" in model:
        pytest.skip(reason="Prompt logprobs does not support FP8")
    num_prompt_logprobs = 5

    prompts = get_chicken_soup_prompts(4)

    monkeypatch.setenv("VLLM_SPYRE_DYNAMO_BACKEND", backend)
    monkeypatch.setenv("VLLM_SPYRE_ENABLE_PROMPT_LOGPROBS", "1")
    tokenizer = AutoTokenizer.from_pretrained(model.name, revision=model.revision)
    llm = LLM(model, tensor_parallel_size=tp_size, tokenizer=tokenizer)

    responses: list[RequestOutput] = llm.generate(
        prompts, sampling_params=SamplingParams(prompt_logprobs=num_prompt_logprobs)
    )

    expected_prompt_logprobs: dict[str, list] = _get_hf_prompt_logprobs(
        model_info=model, prompts=prompts, n=num_prompt_logprobs
    )

    for prompt, response in zip(prompts, responses):
        actual_logprobs = response.prompt_logprobs
        expected_logprobs = expected_prompt_logprobs[prompt]
        _compare_prompt_logprobs(
            expected_logprobs, actual_logprobs, max_different_tokens=1, relative_tolerance=0.15
        )
    force_engine_shutdown(llm)


@pytest.mark.skip
@pytest.mark.cpu
@pytest.mark.decoder
def test_prompt_logprobs_must_be_enabled(monkeypatch: pytest.MonkeyPatch):
    # If prompt logprobs is disabled, requests are rejected
    monkeypatch.setenv("VLLM_SPYRE_ENABLE_PROMPT_LOGPROBS", "0")
    params = SamplingParams(prompt_logprobs=5)

    with pytest.raises(ValueError, match="VLLM_SPYRE_ENABLE_PROMPT_LOGPROBS"):
        SpyrePlatform.validate_request("test-any-prompt", params)


@pytest.mark.skip
@pytest.mark.cpu
@pytest.mark.decoder
def test_prompt_logprobs_not_supported_with_cb(model: str, monkeypatch: pytest.MonkeyPatch):
    # Server shouldn't boot with both prompt logprobs and continuous batching
    # enabled
    monkeypatch.setenv("VLLM_SPYRE_ENABLE_PROMPT_LOGPROBS", "1")
    monkeypatch.setenv("VLLM_SPYRE_USE_CB", "1")

    with pytest.raises(ValueError, match="continuous batching"):
        VllmConfig(
            model_config=ModelConfig(model=model.name, revision=model.revision, task="generate")
        )


@pytest.mark.skip
@pytest.mark.cpu
@pytest.mark.decoder
def test_prompt_logprobs_on_single_requests_only(model: str, monkeypatch: pytest.MonkeyPatch):
    # Only bs=1 is supported for prompt logprobs
    monkeypatch.setenv("VLLM_SPYRE_ENABLE_PROMPT_LOGPROBS", "1")
    monkeypatch.setenv("VLLM_SPYRE_WARMUP_BATCH_SIZES", "2")

    with pytest.raises(ValueError, match="batch size 1"):
        VllmConfig(model_config=ModelConfig(model=model.name, task="generate"))


def _compare_prompt_logprobs(
    expected: list, actual: list, max_different_tokens: int, relative_tolerance: float
):
    # Fuzzy comparison of prompt logprob outputs
    # max_different_tokens is the number of candidate tokens that are allowed to
    # differ at each token in the prompt.
    # relative_tolerance is the tolerance between the expected and actual
    # logprob for each token
    assert len(expected) == len(actual)

    for i in range(len(actual)):
        expected_dict = expected[i]
        actual_dict = actual[i]

        if expected_dict is None and actual_dict is None:
            continue

        expected_token_set = set(expected_dict.keys())
        actual_token_set = set(actual_dict.keys())

        # Check that (most of) the top n tokens are the same
        assert len(expected_token_set - actual_token_set) <= max_different_tokens

        for token, actual_logprob in actual_dict.items():
            # skip tokens not in the expected set
            if token not in expected_dict:
                continue

            expected_logprob = expected_dict[token]

            # 60% tolerance- pretty big difference in results atm
            assert math.isclose(
                expected_logprob["logprob"], actual_logprob.logprob, rel_tol=relative_tolerance
            )


def _get_hf_prompt_logprobs(model_info: ModelInfo, prompts, n) -> dict[str, list]:
    """Get prompt logprobs from HF model directly, including top n candidates
    for each token"""
    tokenizer = AutoTokenizer.from_pretrained(model_info.name, revision=model_info.revision)
    model = AutoModelForCausalLM.from_pretrained(model_info.name, revision=model_info.revision)

    prompt_logprobs = {}
    for prompt in prompts:
        inputs = tokenizer(prompt, return_tensors="pt")
        input_ids = inputs["input_ids"]

        # Get logits (model output before softmax)
        with torch.no_grad():
            outputs = model(**inputs)
            logits = outputs.logits

        # Shift logits and labels so that tokens align with their predicted
        # logprobs
        shifted_logits = logits[:, :-1, :]  # Remove last logit
        shifted_input_ids = input_ids[:, 1:]  # Remove first token (BOS)

        # Get log-softmax over vocabulary
        log_probs = torch.nn.functional.log_softmax(shifted_logits, dim=-1)

        # Get top n logprobs:
        topk_logprobs, topk_indices = torch.topk(log_probs, dim=2, k=n)

        # Gather log-probabilities of the actual prompt logprobs
        token_logprobs = log_probs.gather(2, shifted_input_ids.unsqueeze(-1)).squeeze(-1)

        # Squeeze out batch dimension 1
        token_logprobs = token_logprobs.squeeze()
        topk_logprobs = topk_logprobs.squeeze()
        topk_indices = topk_indices.squeeze()

        # No prompt logprobs for first token
        output_prompt_logprobs = [None]
        for idx in range(len(token_logprobs)):
            logprobs_dict = {}
            # Loop over each token and:
            # Get the set of top N tokens + the actual prompt token
            # (The prompt token may or may not be in the top N)
            # Detokenize and store the token w/ its logprob
            for i, token in enumerate(topk_indices[idx]):
                logprob = topk_logprobs[idx][i]
                text = tokenizer.convert_ids_to_tokens(token.item())
                logprobs_dict[token.item()] = {
                    "decoded_token": text,
                    "logprob": logprob.item(),
                }

            prompt_token = input_ids[0][idx + 1]
            prompt_logprob = token_logprobs[idx]
            decoded_prompt_token = tokenizer.convert_ids_to_tokens(prompt_token.item())
            logprobs_dict[prompt_token.item()] = {
                "decoded_token": decoded_prompt_token,
                "logprob": prompt_logprob.item(),
            }

            output_prompt_logprobs.append(logprobs_dict)

        prompt_logprobs[prompt] = output_prompt_logprobs

    return prompt_logprobs



"""Verification of seeded random sampling to be deterministic

Run `python -m pytest tests/e2e/test_spyre_seed.py`.
"""

import math

import pytest
from output_util import generate_spyre_vllm_output, kwargs_for_mode
from spyre_util import DecodeWarmupShapes, ModelInfo, get_chicken_soup_prompts
from vllm import SamplingParams

sb_mark = pytest.param("sb", marks=pytest.mark.sb, id="sb")
cb_mark = pytest.param("cb", marks=pytest.mark.cb, id="cb")


@pytest.mark.parametrize("temperature", [0.1, 1.0])
@pytest.mark.parametrize("seed", [42])
def test_seed(
    model: ModelInfo,
    temperature: float,
    seed: int,
    max_model_len: int,
    max_num_seqs: int,
    warmup_shapes: DecodeWarmupShapes,
    backend: str,
    mode: str,
    monkeypatch: pytest.MonkeyPatch,
    use_llm_cache,
    runtime_xfail,
) -> None:
    """
    The warmup is based on a single shape. After the warmup,
    output is generated for one request with 5 identical prompts
    using random sampling (non-zero temperature) in combination
    with a seed. The generated output, including text, token ids,
    logprobs is verified to be identical for all 5 sequences.
    """

    if mode in ["cp", "pc"]:
        runtime_xfail(reason="Seed is broken in chunked prefill, please fix me!")

    max_new_tokens = warmup_shapes[0][1]

    prompts = get_chicken_soup_prompts(1) * 5

    vllm_sampling_params = SamplingParams(
        max_tokens=max_new_tokens,
        temperature=temperature,
        logprobs=0,  # return logprobs of generated tokens only
        ignore_eos=True,
        seed=seed,
    )

    vllm_results = generate_spyre_vllm_output(
        model=model,
        prompts=prompts,
        max_model_len=max_model_len,
        sampling_params=vllm_sampling_params,
        tensor_parallel_size=1,
        backend=backend,
        monkeypatch=monkeypatch,
        **kwargs_for_mode(mode, max_num_seqs, warmup_shapes),
    )

    # compare all generated outputs against the first generated output
    for vllm_result in vllm_results:
        assert vllm_result["text"] == vllm_results[0]["text"]

        # compare logprobs for all tokens between
        # the current and the first sequence
        assert len(vllm_result["logprobs"]) == len(vllm_results[0]["logprobs"])
        for token_id, logprob, token_id_0, logprob_0 in zip(
            vllm_result["token_ids"],
            vllm_result["logprobs"],
            vllm_results[0]["token_ids"],
            vllm_results[0]["logprobs"],
        ):
            assert token_id == token_id_0
            assert math.isclose(logprob, logprob_0, rel_tol=0.1)



"""Verification of vLLM output by comparing with HF

Run `python -m pytest tests/e2e/test_spyre_basic.py`.
"""

import asyncio
from typing import cast
from unittest.mock import patch

import pytest
import requests
from llm_cache import get_cached_llm
from output_util import compare_results, extract_output, generate_hf_output, setup_golden_token
from pytest_mock.plugin import MockerFixture
from spyre_util import (
    ModelInfo,
    RemoteOpenAIServer,
    get_chicken_soup_prompts,
    get_longer_chicken_soup_prompts,
)
from vllm import LLM, SamplingParams

from tests.scheduling_utils import random_prompt
from vllm_spyre.v1.worker.spyre_model_runner import SamplingForwardInputs


def get_model_runner(cp_model: LLM):
    return (
        cp_model.llm_engine.engine_core.engine_core.model_executor.driver_worker.worker.model_runner
    )


chicken_soup_prompts = get_longer_chicken_soup_prompts(4)

# NOTE: considering granite 3.3 tokenizer
# Should have 51 tokens
prompt_51 = get_chicken_soup_prompts(1)[0]
# Should have 95 tokens
prompt_95 = chicken_soup_prompts[0]
# Should have 251
prompt_251 = chicken_soup_prompts[0] + chicken_soup_prompts[1] + chicken_soup_prompts[2]
# Should have 260 tokens
prompt_260 = chicken_soup_prompts[0] + chicken_soup_prompts[2] + chicken_soup_prompts[3]

USE_CASES = {
    # Case I - Prompt fits in a single chunk
    "case_Ia": (prompt_95, 128, 1, 0),
    # Case I - Prompt fits in a single chunk with left padding
    "case_Ib": (prompt_51, 128, 1, 64),
    # Case II - Has left padding
    "case_II": (prompt_260, 128, 3, 64),
    # Case III again - no padding
    "case_III": (prompt_251, 128, 2, 0),
}


@pytest.mark.chunked_prefill
@pytest.mark.parametrize("use_case", list(USE_CASES.keys()))
def test_chunked_prefill_correctness(
    model: ModelInfo,
    backend: str,
    max_num_seqs: int,
    max_model_len: int,
    monkeypatch: pytest.MonkeyPatch,
    mocker: MockerFixture,
    use_case: str,
    use_llm_cache,
) -> None:
    """
    Minimal test to check if vllm-spyre activate code for chunked prefill for
    a prompt greater than the chunk size
    """

    (prompt, chunk_size, expected_chunk_count, expected_left_padding) = USE_CASES[use_case]
    max_new_tokens = 8

    hf_outputs = generate_hf_output(
        model=model,
        prompts=[prompt],
        max_new_tokens=max_new_tokens,
        ignore_eos=True,
    )
    ### NB: May not be guaranteed to be set
    monkeypatch.setenv("VLLM_ENABLE_V1_MULTIPROCESSING", "0")

    cp_model = get_cached_llm(
        model=model,
        max_model_len=max_model_len,
        tensor_parallel_size=1,
        backend=backend,
        monkeypatch=monkeypatch,
        max_num_seqs=max_num_seqs,
        use_cb=True,
        max_num_batched_tokens=chunk_size,
    )
    model_runner = get_model_runner(cp_model)

    sampling_params = SamplingParams(
        max_tokens=max_new_tokens, temperature=0, logprobs=0, ignore_eos=True
    )
    git_sampling_params = setup_golden_token(model, sampling_params, hf_outputs)

    _prepare_chunked_prefill = model_runner._prepare_chunked_prefill
    records = []

    def wrapper(self, *args, **kwargs):
        model_input = _prepare_chunked_prefill(self, *args, **kwargs)
        records.append(model_input)
        return model_input

    with patch.object(model_runner, "_prepare_chunked_prefill", wraps=wrapper) as spy:
        results = cp_model.generate(prompt, git_sampling_params)
        vllm_results = [extract_output(results[0])]

        for r in records:
            model_input = cast(SamplingForwardInputs, r)
            # Must be a single value
            left_padding = model_input.left_padded_prompt_mask[0].item()
            assert left_padding == expected_left_padding

    # Validate if the prefill was chunked
    spy.assert_called()
    assert spy.call_count == expected_chunk_count

    # Validate output
    compare_results(
        model=model,
        tensor_parallel_size=1,
        backend=backend,
        vllm_results=vllm_results,
        hf_results=hf_outputs,
        prompts=[prompt],
    )


@pytest.mark.parametrize("mode", [pytest.param("pc", marks=pytest.mark.prefix_caching, id="pc")])
@pytest.mark.asyncio
async def test_chunked_prefill_kv_cache_stats(
    remote_openai_server: RemoteOpenAIServer,
    model,
    warmup_shapes,
    backend,
    tp_size,
    mode,
    max_num_seqs,
    max_model_len,
):
    # Test that vllm metrics include prefix caching data
    client = remote_openai_server.get_async_client()

    prompt = random_prompt(
        model=model,
        seed=0,
        length=max_model_len // 2,  # try to span multiple chunks
    )

    # Start a bunch of duplicate requests so that we hit prefix cache:
    request_tasks: list[asyncio.Task] = []
    # 2x max batch size so that there's some queueing
    for _ in range(2 * max_num_seqs):
        request_tasks.append(
            asyncio.create_task(
                client.completions.create(model=model.name, prompt=prompt, max_tokens=5)
            )
        )

    # Wait for the first request to finish to ensure that requests are
    # processing
    requests_finished = 0
    while requests_finished < 1:
        done, pending = await asyncio.wait(request_tasks, return_when=asyncio.FIRST_COMPLETED)
        requests_finished += len(done)
        request_tasks = pending

    # Now that requests are processing, check metrics for KV cache usage.
    # This must be done while requests are in-flight, once they finish this
    # gauge will be 0.
    metrics = get_metrics(remote_openai_server)
    kv_cache_usage = get_metric_value(metrics, "vllm:kv_cache_usage_perc")
    assert kv_cache_usage > 0

    # Wait for at least 2 requests to finish to ensure the prefix cache was used
    while requests_finished < 2:
        done, pending = await asyncio.wait(request_tasks, return_when=asyncio.FIRST_COMPLETED)
        requests_finished += len(done)
        request_tasks = pending

    # Cancel the rest
    for task in request_tasks:
        task.cancel()

    # Check the prefix cache counters
    # vLLM should be reporting these counters based on the last 1000 requests
    # (as of v0.12.0)
    metrics = get_metrics(remote_openai_server)
    total_tokens = get_metric_value(metrics, "vllm:prefix_cache_queries_total")
    hit_tokens = get_metric_value(metrics, "vllm:prefix_cache_hits_total")

    # Prefix cache rate won't be 100% because of chunking, but we should still
    # hit _some_ cache
    assert hit_tokens / total_tokens > 0.1


def get_metrics(remote_openai_server: RemoteOpenAIServer) -> list[str]:
    metrics_response = requests.get(f"http://localhost:{remote_openai_server.port}/metrics")
    assert metrics_response.status_code == 200
    metrics = metrics_response.text
    return metrics.splitlines()


def get_metric_value(metrics: list[str], metric_name: str) -> float:
    metric_line = [line for line in metrics if line.startswith(metric_name)][0]
    return float(metric_line.split(" ")[-1])



"""Verification of vLLM output by comparing with HF

Run `python -m pytest tests/e2e/test_spyre_max_new_tokens.py`.
"""

import pytest
from output_util import validate_vllm_vs_hf_output, kwargs_for_mode
from spyre_util import DecodeWarmupShapes, ModelInfo, get_chicken_soup_prompts
from vllm import SamplingParams

sb_mark = pytest.param("sb", marks=pytest.mark.sb, id="sb")
cb_mark = pytest.param("cb", marks=pytest.mark.cb, id="cb")
cp_mark = pytest.param("cp", marks=pytest.mark.chunked_prefill, id="cp")


@pytest.mark.parametrize("stop_last", [True, False])
@pytest.mark.parametrize("mode", [sb_mark, cb_mark, cp_mark])
def test_output(
    model: ModelInfo,
    stop_last: bool,
    max_model_len: int,
    max_num_seqs: int,
    warmup_shapes: DecodeWarmupShapes,
    backend: str,
    mode: str,
    monkeypatch: pytest.MonkeyPatch,
    use_llm_cache,
) -> None:
    """
    Checks that `max_tokens` parameter of `SamplingParams` works correctly
    For each batch, one prompt has max_tokens set to 1 and the others don't.
    This checks that the correct request has only a single output token, while
    the others are not affected.
    """

    prompts = get_chicken_soup_prompts(4)

    max_new_tokens_long = 6
    max_new_tokens_early_stop = 1

    vllm_sampling_params_normal = SamplingParams(
        max_tokens=max_new_tokens_long,
        temperature=0,
        logprobs=0,  # return logprobs of generated tokens only
        ignore_eos=False,
    )

    vllm_sampling_params_early_stop = SamplingParams(
        max_tokens=max_new_tokens_early_stop,
        temperature=0,
        logprobs=0,  # return logprobs of generated tokens only
        ignore_eos=False,
    )

    vllm_sampling_params = [vllm_sampling_params_normal.clone() for _ in range(3)]
    hf_max_new_tokens = [max_new_tokens_long] * 3

    # stop last or first sequence in batch early
    if stop_last:
        vllm_sampling_params = vllm_sampling_params + [vllm_sampling_params_early_stop]
        hf_max_new_tokens = hf_max_new_tokens + [max_new_tokens_early_stop]
    else:
        vllm_sampling_params = [vllm_sampling_params_early_stop] + vllm_sampling_params
        hf_max_new_tokens = [max_new_tokens_early_stop] + hf_max_new_tokens

    validate_vllm_vs_hf_output(
        model=model,
        prompts=prompts,
        sampling_params=vllm_sampling_params,
        tensor_parallel_size=1,
        backend=backend,
        monkeypatch=monkeypatch,
        max_new_tokens=hf_max_new_tokens,
        max_model_len=max_model_len,
        **kwargs_for_mode(mode, max_num_seqs, warmup_shapes),
    )



import os

import pytest
from vllm.v1.core.block_pool import BlockPool
from vllm.v1.core.sched.output import SchedulerOutput
from vllm.v1.core.single_type_kv_cache_manager import FullAttentionManager

from vllm_spyre.compat_utils import dataclass_fields, has_argument

pytestmark = pytest.mark.compat

VLLM_VERSION = os.getenv("TEST_VLLM_VERSION", "default")


def test_pin_memory_available():
    if VLLM_VERSION == "vLLM:lowest":
        try:
            from vllm.utils import is_pin_memory_available  # # noqa #ty: ignore
            from vllm.utils import make_tensor_with_pad  # # noqa #ty: ignore
            from vllm.utils import init_cached_hf_modules  # # noqa #ty: ignore
        except ImportError as e:
            raise AssertionError(
                "remove backwards compatibility imports for "
                "is_pin_memory_available, "
                "make_tensor_with_pad and init_cached_hf_modules"
            ) from e


def test_multi_modal_cache_stats():
    if VLLM_VERSION == "vLLM:lowest":
        # If this import succeeds then remove the backwards compatibility type
        # def for MultiModalCacheStats
        with pytest.raises(ImportError):
            from vllm.v1.metrics.stats import MultiModalCacheStats  # # noqa #ty: ignore


def test_v0_worker_base():
    if VLLM_VERSION == "vLLM:lowest":
        try:
            from vllm.worker.worker_base import WorkerBase  # # noqa #ty: ignore
        except ImportError as e:
            raise AssertionError(
                "remove the backwards compatibility code from the SpyreWorker initializer"
            ) from e


def test_structured_output_request_ids():
    if VLLM_VERSION == "vLLM:lowest":
        # Can remove "structured_output_request_ids" and "grammar_bitmask"
        # from backwards compat
        assert "structured_output_request_ids" in dataclass_fields(SchedulerOutput)


def test_hash_block_size():
    if VLLM_VERSION == "vLLM:lowest":
        # Can supply `hash_block_size` everywhere, this was added in 0.12.0
        assert not has_argument(BlockPool, "hash_block_size")


def test_alignment_tokens():
    if VLLM_VERSION == "vLLM:lowest":
        # Can supply `alignment_tokens` everywhere, this was added in 0.12.0
        assert not has_argument(FullAttentionManager.find_longest_cache_hit, "alignment_tokens")


def test_argparse_utils():
    if VLLM_VERSION == "vLLM:lowest":
        try:
            from vllm.utils import FlexibleArgumentParser  # noqa
        except ImportError as e:
            raise AssertionError(
                "Fix backward compatible imports of "
                "FlexibleArgumentParser which is no longer required"
            ) from e


def test_pooler_api():
    if VLLM_VERSION == "vLLM:lowest":
        try:
            from vllm.model_executor.layers.pooler import ClassifierPooler, Pooler  # noqa
        except ImportError as e:
            raise AssertionError(
                "Backwards compatibility code for old pooler API "
                "ClassifierPooler no longer required, related to vLLM PR #31973"
            ) from e


def test_set_random_seed():
    if VLLM_VERSION == "vLLM:lowest":
        try:
            from vllm.model_executor import set_random_seed  # noqa
        except ImportError as e:
            raise AssertionError(
                "Backwards compatibility code for set_random_seed import no longer required"
            ) from e


def test_enable_caching():
    if VLLM_VERSION == "vLLM:lowest":
        # Can supply enable_caching everywhere, added in v0.14.0
        assert not has_argument(FullAttentionManager.__init__, "enable_caching"), (
            "Backwards compatibility code for enable_caching parameter "
            "in FullAttentionManager no longer required"
        )


def test_pooling_metadata_build_cursor():
    if VLLM_VERSION == "vLLM:lowest":
        from vllm.v1.pool.metadata import PoolingMetadata

        assert has_argument(PoolingMetadata.build_pooling_cursor, "num_scheduled_tokens"), (
            "Backwards compatibility code for num_scheduled_tokens parameter "
            "in PoolingMetadata.build_pooling_cursor no longer required "
        )


def test_allocate_new_computed_blocks():
    if VLLM_VERSION == "vLLM:lowest":
        # allocate_new_computed_blocks was added in v0.14.0
        # When save_new_computed_blocks no longer exists, remove the
        # try/except compatibility code in spyre_model_runner.py
        assert hasattr(FullAttentionManager, "save_new_computed_blocks"), (
            "Backwards compatibility code for save_new_computed_blocks "
            "in FullAttentionManager no longer required, can use "
            "allocate_new_computed_blocks everywhere"
        )



from collections import defaultdict

import pytest
from llm_cache import get_cached_llm
from spyre_util import ModelInfo
from vllm import SamplingParams

pytestmark = [pytest.mark.full_model, pytest.mark.other_e2e]

sb_mark = pytest.param("sb", marks=pytest.mark.sb, id="sb")
cb_mark = pytest.param("cb", marks=pytest.mark.cb, id="cb")
cp_mark = pytest.param("cp", marks=pytest.mark.chunked_prefill, id="cp")


def test_spyre_batch1_temperature(
    model: ModelInfo, backend, monkeypatch, use_llm_cache, warmup_shapes
):
    spyre_model = get_cached_llm(
        model=model,
        max_model_len=128,
        tensor_parallel_size=1,
        backend=backend,
        monkeypatch=monkeypatch,
        warmup_shapes=warmup_shapes,
    )

    prompt = "The capital of the United Kingdom is"
    params1 = SamplingParams(temperature=0.0, seed=8780, max_tokens=20)
    params2 = SamplingParams(temperature=0.5, seed=8780, max_tokens=20)
    params3 = SamplingParams(temperature=1.0, seed=8780, max_tokens=20)

    output1 = spyre_model.generate(prompt, params1)[0]
    output2 = spyre_model.generate(prompt, params2)[0]
    output3 = spyre_model.generate(prompt, params3)[0]

    assert output1.outputs[0].text != output2.outputs[0].text
    assert output2.outputs[0].text != output3.outputs[0].text


def test_spyre_batch1_max_tokens(
    model: ModelInfo, backend, monkeypatch, use_llm_cache, warmup_shapes
):
    spyre_model = get_cached_llm(
        model=model,
        max_model_len=128,
        tensor_parallel_size=1,
        backend=backend,
        monkeypatch=monkeypatch,
        warmup_shapes=warmup_shapes,
    )

    prompt = "Count to twenty"
    params1 = SamplingParams(temperature=0, seed=8780, max_tokens=15)
    params2 = SamplingParams(temperature=0, seed=8780)

    output1 = spyre_model.generate(prompt, params1)[0]
    output2 = spyre_model.generate(prompt, params2)[0]

    assert len(output1.outputs[0].token_ids) == 15
    assert len(output2.outputs[0].token_ids) > 15


@pytest.mark.xfail(reason="Failing currently because of output mismatch")
def test_spyre_batch1_stop_sequence(
    model: ModelInfo, backend, monkeypatch, use_llm_cache, warmup_shapes
):
    spyre_model = get_cached_llm(
        model=model,
        max_model_len=128,
        tensor_parallel_size=1,
        backend=backend,
        monkeypatch=monkeypatch,
        warmup_shapes=warmup_shapes,
    )
    stop_str = "train"
    prompt = "The best way to travel from Paris to Berlim is by "

    params1 = SamplingParams(stop=[stop_str], max_tokens=20, seed=8780)
    params2 = SamplingParams(max_tokens=20, seed=8780)

    output1 = spyre_model.generate(prompt, params1)[0]
    output2 = spyre_model.generate(prompt, params2)[0]

    assert stop_str not in output1.outputs[0].text
    assert output1.outputs[0].finish_reason == "stop"
    assert output2.outputs[0].finish_reason != "stop"


def max_repetitions(output):
    histo = defaultdict(int)
    for token in output.outputs[0].token_ids:
        histo[token] += 1

    return max(histo.values())


def test_spyre_batch1_presence_penalty(
    model: ModelInfo, backend, monkeypatch, use_llm_cache, warmup_shapes
):
    spyre_model = get_cached_llm(
        model=model,
        max_model_len=128,
        tensor_parallel_size=1,
        backend=backend,
        monkeypatch=monkeypatch,
        warmup_shapes=warmup_shapes,
    )
    prompt = "REPEAT OVER AND OVER AGAIN THE MINIMUM TIMES POSSIBLE: one one one one one"

    param1 = SamplingParams(presence_penalty=2.0, seed=8780, max_tokens=20)
    param2 = SamplingParams(presence_penalty=-2.0, seed=8780, max_tokens=20)

    with_penalty = spyre_model.generate(prompt, param1)[0]
    no_penalty = spyre_model.generate(prompt, param2)[0]

    with_penalty_max = max_repetitions(with_penalty)
    no_penalty_max = max_repetitions(no_penalty)

    assert no_penalty_max > 1
    assert no_penalty_max > with_penalty_max


def test_spyre_batch1_frequency_penalty(
    model: ModelInfo, backend, monkeypatch, use_llm_cache, warmup_shapes
):
    spyre_model = get_cached_llm(
        model=model,
        max_model_len=128,
        tensor_parallel_size=1,
        backend=backend,
        monkeypatch=monkeypatch,
        warmup_shapes=warmup_shapes,
    )

    prompt = "repeat the word hi ten times:"

    param1 = SamplingParams(frequency_penalty=2.0, seed=8780, max_tokens=20)
    param2 = SamplingParams(frequency_penalty=-2.0, seed=8780, max_tokens=20)

    with_penalty = spyre_model.generate(prompt, param1)[0]
    no_penalty = spyre_model.generate(prompt, param2)[0]

    with_penalty_max = max_repetitions(with_penalty)
    no_penalty_max = max_repetitions(no_penalty)
    assert no_penalty_max > 1
    assert no_penalty_max > with_penalty_max


def test_spyre_batch1_n_generations(
    model: ModelInfo, backend, monkeypatch, use_llm_cache, warmup_shapes
):
    spyre_model = get_cached_llm(
        model=model,
        max_model_len=128,
        tensor_parallel_size=1,
        backend=backend,
        monkeypatch=monkeypatch,
        warmup_shapes=warmup_shapes,
    )
    prompt = "The three most popular sports in the world are: "

    params = SamplingParams(n=3, seed=8780, max_tokens=20)

    output = spyre_model.generate(prompt, params)[0]

    assert len(output.outputs) == 3
    assert output.outputs[0].text != output.outputs[1].text
    assert output.outputs[1].text != output.outputs[2].text


def token_diversity(spyre_model, prompt, params, n_experiments):
    tokens = []

    outputs = spyre_model.generate([prompt] * n_experiments, params, use_tqdm=False)
    for output in outputs:
        tokens.extend(output.outputs[0].token_ids)

    return len(set(tokens))


def test_spyre_batch1_top_p(model: ModelInfo, backend, monkeypatch, use_llm_cache, warmup_shapes):
    spyre_model = get_cached_llm(
        model=model,
        max_model_len=128,
        tensor_parallel_size=1,
        backend=backend,
        monkeypatch=monkeypatch,
        warmup_shapes=warmup_shapes,
    )
    prompt = "The first three letters of the alphabet are"
    params1 = SamplingParams(top_p=0.01, temperature=1, max_tokens=10)
    params2 = SamplingParams(temperature=1, max_tokens=10)

    token_div1 = token_diversity(spyre_model, prompt, params1, 10)
    token_div2 = token_diversity(spyre_model, prompt, params2, 10)
    assert token_div1 < token_div2


def test_spyre_batch1_top_k(model: ModelInfo, backend, monkeypatch, use_llm_cache, warmup_shapes):
    spyre_model = get_cached_llm(
        model=model,
        max_model_len=128,
        tensor_parallel_size=1,
        backend=backend,
        monkeypatch=monkeypatch,
        warmup_shapes=warmup_shapes,
    )
    prompt = "The opposite of hot is"
    params1 = SamplingParams(temperature=1, top_k=1, max_tokens=5)
    params2 = SamplingParams(temperature=1, max_tokens=5)

    token_div1 = token_diversity(spyre_model, prompt, params1, 10)
    token_div2 = token_diversity(spyre_model, prompt, params2, 10)
    assert token_div1 < token_div2


@pytest.mark.parametrize("mode", [sb_mark, cb_mark, cp_mark])
def test_spyre_batch1_logit_bias(
    model: ModelInfo,
    backend,
    monkeypatch,
    use_llm_cache,
    warmup_shapes,
    max_model_len,
    max_num_seqs,
    mode: str,
):
    spyre_model = get_cached_llm(
        model=model,
        max_model_len=max_model_len,
        tensor_parallel_size=1,
        backend=backend,
        monkeypatch=monkeypatch,
        warmup_shapes=warmup_shapes if mode == "sb" else None,
        max_num_seqs=max_num_seqs if mode == "cb" or mode == "cp" else None,
        max_num_batched_tokens=128 if mode == "cp" else None,
        use_cb=mode == "cb" or mode == "cp",
    )
    tokenizer = spyre_model.get_tokenizer()
    banned_word = "train"
    forced_word = "plane"

    banned_ids = tokenizer.encode(banned_word, add_special_tokens=False)
    forced_ids = tokenizer.encode(forced_word, add_special_tokens=False)

    banned_word_id = banned_ids[0]
    forced_word_id = forced_ids[0]

    prompt = "The fastest way to travel between continents is by "
    params1 = SamplingParams(
        temperature=0,
        max_tokens=5,
        seed=8780,
        logit_bias={
            banned_word_id: -100,
            forced_word_id: 100,
        },
    )
    params2 = SamplingParams(temperature=0, seed=8780, max_tokens=5)

    output = spyre_model.generate([prompt, prompt], [params1, params2])

    assert banned_word not in output[0].outputs[0].text.lower()
    assert forced_word in output[0].outputs[0].text.lower()

    assert output[0].outputs[0].text != output[1].outputs[0].text


@pytest.mark.parametrize("mode", [sb_mark, cb_mark, cp_mark])
def test_spyre_batch1_min_tokens(
    model: ModelInfo,
    backend,
    monkeypatch,
    use_llm_cache,
    max_model_len,
    max_num_seqs,
    warmup_shapes,
    mode: str,
):
    spyre_model = get_cached_llm(
        model=model,
        max_model_len=max_model_len,
        tensor_parallel_size=1,
        backend=backend,
        monkeypatch=monkeypatch,
        warmup_shapes=warmup_shapes if mode == "sb" else None,
        max_num_seqs=max_num_seqs if mode == "cb" or mode == "cp" else None,
        max_num_batched_tokens=128 if mode == "cp" else None,
        use_cb=mode == "cb" or mode == "cp",
    )
    prompt = "What is the capital of the USA?"
    tokenizer = spyre_model.get_tokenizer()
    eos_id = tokenizer.eos_token_id

    params1 = SamplingParams(min_tokens=10, logit_bias={eos_id: 1000}, seed=8780, max_tokens=20)
    params2 = SamplingParams(seed=8780, logit_bias={eos_id: 1000}, max_tokens=20)

    output = spyre_model.generate([prompt] * 2, [params1, params2])

    # Logits bias should force eos token appears, then we check if
    # after min tokens reached the logits processor is properly
    # cleared. Therefore token count shall be 10 + 1
    # (min_tokens + eos_token_id)
    assert len(output[0].outputs[0].token_ids) == 11
    assert len(output[1].outputs[0].token_ids) == 1


def test_spyre_batch1_ignore_eos(
    model: ModelInfo, backend, monkeypatch, use_llm_cache, warmup_shapes
):
    spyre_model = get_cached_llm(
        model=model,
        max_model_len=128,
        tensor_parallel_size=1,
        backend=backend,
        monkeypatch=monkeypatch,
        warmup_shapes=warmup_shapes,
    )
    tokenizer = spyre_model.get_tokenizer()
    eos_id = tokenizer.eos_token_id
    prompt = "This is the end of the story"

    params1 = SamplingParams(ignore_eos=True, logit_bias={eos_id: 50}, seed=8780, max_tokens=20)
    params2 = SamplingParams(ignore_eos=False, logit_bias={eos_id: 50}, seed=8780, max_tokens=20)

    output1 = spyre_model.generate(prompt, params1)[0]
    output2 = spyre_model.generate(prompt, params2)[0]

    assert len(output1.outputs[0].token_ids) == 20
    assert len(output2.outputs[0].token_ids) != len(output1.outputs[0].token_ids)

    assert output1.outputs[0].finish_reason == "length"
    assert output2.outputs[0].finish_reason != "length"


@pytest.mark.parametrize("mode", [sb_mark, cb_mark, cp_mark])
def test_spyre_batch1_min_p(
    model: ModelInfo,
    backend,
    monkeypatch,
    use_llm_cache,
    max_model_len,
    max_num_seqs,
    warmup_shapes,
    mode: str,
):
    spyre_model = get_cached_llm(
        model=model,
        max_model_len=max_model_len,
        tensor_parallel_size=1,
        backend=backend,
        monkeypatch=monkeypatch,
        warmup_shapes=warmup_shapes if mode == "sb" else None,
        max_num_seqs=max_num_seqs if mode == "cb" or mode == "cp" else None,
        max_num_batched_tokens=128 if mode == "cp" else None,
        use_cb=mode == "cb" or mode == "cp",
    )
    prompt = "The opposite of black is"
    params1 = SamplingParams(min_p=0.5, temperature=1, max_tokens=5)
    params2 = SamplingParams(temperature=1, max_tokens=5)

    token_div1 = token_diversity(spyre_model, prompt, params1, 10)
    token_div2 = token_diversity(spyre_model, prompt, params2, 10)

    assert token_div1 < token_div2


@pytest.mark.xfail(reason="Failing currently because of output mismatch")
def test_spyre_batch1_bad_words(
    model: ModelInfo, backend, monkeypatch, use_llm_cache, warmup_shapes
):
    spyre_model = get_cached_llm(
        model=model,
        max_model_len=128,
        tensor_parallel_size=1,
        backend=backend,
        monkeypatch=monkeypatch,
        warmup_shapes=warmup_shapes,
    )
    prompt = "The capital of France is"
    params1 = SamplingParams(
        max_tokens=5, temperature=0, seed=8780, bad_words=[" Paris", " Parisi", " France"]
    )
    params2 = SamplingParams(max_tokens=5, seed=8780, temperature=0)

    output1 = spyre_model.generate(prompt, params1)[0]
    output2 = spyre_model.generate(prompt, params2)[0]

    assert "Paris" not in output1.outputs[0].text
    assert "France" not in output1.outputs[0].text
    assert output1.outputs[0].text != output2.outputs[0].text


def test_spyre_batch1_detokenize(
    model: ModelInfo, backend, monkeypatch, use_llm_cache, warmup_shapes
):
    spyre_model = get_cached_llm(
        model=model,
        max_model_len=128,
        tensor_parallel_size=1,
        backend=backend,
        monkeypatch=monkeypatch,
        warmup_shapes=warmup_shapes,
    )
    prompt = "Hello, world!"
    params = SamplingParams(max_tokens=5, seed=8780, temperature=0, detokenize=False)
    output = spyre_model.generate(prompt, params)[0]

    assert output.outputs[0].text == ""
    assert len(output.outputs[0].token_ids) > 0


def test_spyre_batch1_logprobs(
    model: ModelInfo, backend, monkeypatch, use_llm_cache, warmup_shapes
):
    spyre_model = get_cached_llm(
        model=model,
        max_model_len=128,
        tensor_parallel_size=1,
        backend=backend,
        monkeypatch=monkeypatch,
        warmup_shapes=warmup_shapes,
    )
    num_logprobs = 5
    prompt = "The sky is"
    params = SamplingParams(max_tokens=5, seed=8780, temperature=0, logprobs=num_logprobs)
    output = spyre_model.generate(prompt, params)[0]

    completion_output = output.outputs[0]

    assert completion_output.logprobs is not None
    assert len(completion_output.logprobs) == len(completion_output.token_ids)
    assert len(completion_output.logprobs[0]) == num_logprobs



"""Tests the chunked prefill logic in the model runner

These tests do not involve the scheduler or engine, they isolate the testing to
the padding and chunking logic in the model runner itself. Theyare designed to
run on cpu only.

These tests all assume a chunk size of 128 to keep the test runtime overhead
low.
"""

from dataclasses import dataclass, field

import pytest
from llm_cache import get_cached_engine
from spyre_util import ModelInfo
from vllm.v1.core.sched.output import CachedRequestData, NewRequestData, SchedulerOutput
from vllm.v1.engine.core import EngineCore
from vllm.v1.outputs import ModelRunnerOutput
from vllm.v1.request import Request, SamplingParams

from vllm_spyre.platform import SpyrePlatform
from vllm_spyre.v1.worker.spyre_model_runner import SpyreModelRunner
from vllm_spyre.v1.worker.spyre_worker import _get_extra_args


########## Assuming that we have:
@dataclass
class ChunkedPrefillModelRunnerOutput(ModelRunnerOutput):
    # Current TKV for each request
    # request_tkvs: dict[str, int] = field(default_factory=dict)
    # JK! We will just return tkv = max(request_tkvs)

    # Number of computed tokens for each request (may be smaller than the number
    # of scheduled tokens)
    computed_tokens: dict[str, int] = field(default_factory=dict)
    # (not implemented yet- maybe will be?)

    # Free KV cache blocks left to use
    n_free_blocks: int = 0
    # Already part of CB model runner output


########## End assumptions


def default_test_params(test_func):
    """Sets params for the tests in this file"""
    test_func = pytest.mark.parametrize(
        "max_model_len", [512], ids=lambda val: f"max_model_len({val})"
    )(test_func)
    test_func = pytest.mark.parametrize(
        "max_num_seqs", [2], ids=lambda val: f"max_num_seqs({val})"
    )(test_func)

    test_func = pytest.mark.parametrize(
        "max_num_batched_tokens", [128], ids=lambda val: f"max_num_batched_tokens({val})"
    )(test_func)

    return test_func


def get_cpu_model_runner(
    model: ModelInfo,
    max_model_len: int,
    max_num_seqs: int,
    max_num_batched_tokens: int,
    monkeypatch: pytest.MonkeyPatch,
) -> SpyreModelRunner:
    # TODO: Need to add chunked prefill mode + params to get_cached_engine
    engine_core: EngineCore = get_cached_engine(
        model=model,
        max_model_len=max_model_len,
        max_num_seqs=max_num_seqs,
        max_num_batched_tokens=max_num_batched_tokens,
        available_blocks=None,
        backend="eager",
        monkeypatch=monkeypatch,
    )

    # NB: Only works because this engine is run with no multiprocessing and TP=1
    runner: SpyreModelRunner = engine_core.model_executor.driver_worker.worker.model_runner

    # Clean things up, this isn't a fixture that cleans up after previous tests
    if runner.requests:
        for request_id in list(runner.requests):
            # aborting one-by-one should be less error prone for the runner
            abort_sched = make_scheduler_output(
                scheduled_new_reqs=[], finished_req_ids={request_id}
            )
            runner.execute_model(abort_sched)

    return runner


def make_cached_request_data(req_id_to_computed_tokens) -> CachedRequestData:
    cached_request_data = CachedRequestData.make_empty()
    cached_request_data.req_ids = list(req_id_to_computed_tokens.keys())
    cached_request_data.num_computed_tokens = list(req_id_to_computed_tokens.values())
    return cached_request_data


def make_scheduler_output(
    scheduled_new_reqs: list[NewRequestData],
    scheduled_cached_reqs: CachedRequestData = None,
    num_scheduled_tokens: dict[str, int] = None,
    finished_req_ids: set[str] = None,
) -> SchedulerOutput:
    total_tokens = sum(num_scheduled_tokens.values()) if num_scheduled_tokens else 0
    if scheduled_cached_reqs is None:
        scheduled_cached_reqs = CachedRequestData.make_empty()
    if num_scheduled_tokens is None:
        num_scheduled_tokens = {}
    if finished_req_ids is None:
        finished_req_ids = set()

    extra_args = _get_extra_args()
    return SchedulerOutput(
        scheduled_new_reqs=scheduled_new_reqs,
        scheduled_cached_reqs=scheduled_cached_reqs,
        num_scheduled_tokens=num_scheduled_tokens,
        total_num_scheduled_tokens=total_tokens,
        scheduled_spec_decode_tokens={},
        scheduled_encoder_inputs={},
        num_common_prefix_blocks=[],
        finished_req_ids=finished_req_ids,
        kv_connector_metadata=None,
        **extra_args,
    )


def make_new_request_data(req_id, prompt_len):
    req = Request(
        request_id=req_id,
        prompt_token_ids=[42] * prompt_len,
        sampling_params=SamplingParams(),
        pooling_params=None,
        eos_token_id=None,
    )
    return NewRequestData.from_request(req, block_ids=[])


@pytest.mark.cpu
@pytest.mark.chunked_prefill
@default_test_params
def test_single_block_chunked_prefill(
    model: ModelInfo,
    max_model_len: int,
    max_num_seqs: int,
    max_num_batched_tokens: int,
    monkeypatch: pytest.MonkeyPatch,
):
    """A request that fits within a single block should be prefilled in one
    step and should not be padded out to the chunk boundary on decode"""
    runner: SpyreModelRunner = get_cpu_model_runner(
        model=model,
        max_model_len=max_model_len,
        max_num_seqs=max_num_seqs,
        max_num_batched_tokens=max_num_batched_tokens,
        monkeypatch=monkeypatch,
    )

    req_id = "1"
    # Sub-block prompt
    prompt_len = SpyrePlatform.get_block_size() - 12
    new_req_data = make_new_request_data(req_id, prompt_len)

    scheduler_output = make_scheduler_output(
        scheduled_new_reqs=[new_req_data], num_scheduled_tokens={req_id: prompt_len}
    )

    output = runner.execute_model(scheduler_output)

    # one output token from prefill
    assert len(output.sampled_token_ids[0]) == 1
    # tkv is prompt_len
    assert output.tkv == prompt_len

    # One decode pass to ensure no extra padding shenanigans
    cached_req_data = make_cached_request_data({req_id: prompt_len})
    scheduler_output = make_scheduler_output(
        scheduled_new_reqs=[],
        scheduled_cached_reqs=cached_req_data,
        num_scheduled_tokens={req_id: 1},
    )
    output = runner.execute_model(scheduler_output)
    assert len(output.sampled_token_ids[0]) == 1
    # No extra block or chunk padding
    assert output.tkv == prompt_len + 1


@pytest.mark.cpu
@pytest.mark.chunked_prefill
@default_test_params
def test_multi_chunk_padded_prefill(
    model: ModelInfo,
    max_model_len: int,
    max_num_seqs: int,
    max_num_batched_tokens: int,
    monkeypatch: pytest.MonkeyPatch,
):
    """A request that's longer than a chunk is split into multiple chunks, and
    left-padded only with full size blocks to the end of the last chunk boundary
    """
    runner: SpyreModelRunner = get_cpu_model_runner(
        model=model,
        max_model_len=max_model_len,
        max_num_seqs=max_num_seqs,
        max_num_batched_tokens=max_num_batched_tokens,
        monkeypatch=monkeypatch,
    )
    block_size = SpyrePlatform.get_block_size()

    req_id = "1"
    # Slightly longer than one chunk, so we'll need full left-block padding
    prompt_len = max_num_batched_tokens + 20
    new_req_data = make_new_request_data(req_id, prompt_len)

    # Scheduler will give first chunk
    scheduler_output = make_scheduler_output(
        scheduled_new_reqs=[new_req_data], num_scheduled_tokens={req_id: max_num_batched_tokens}
    )
    output = runner.execute_model(scheduler_output)

    # no output tokens
    assert len(output.sampled_token_ids) == 0
    # Since this request only has one partial block in the second chunk but it
    # needs to be padded all the way to the last block of the second chunk, only
    # a single block of tokens will actually be processed in the first chunk.
    # 🌶️🌶️🌶️ We probably need to be able to pass back the number of tokens that
    # we actually processed so that the scheduler has an accurate count of
    # remaining prompt tokens.
    # TODO: Uncomment when supported
    # assert output.computed_tokens[req_id] == 64
    computed_tokens = {req_id: block_size}

    # Scheduler schedules remaining 148 - 64 = 84 tokens
    cached_req_data = make_cached_request_data(computed_tokens)
    scheduler_output = make_scheduler_output(
        scheduled_new_reqs=[],
        scheduled_cached_reqs=cached_req_data,
        num_scheduled_tokens={req_id: prompt_len - block_size},
    )
    output = runner.execute_model(scheduler_output)

    # Should be one output token now
    assert len(output.sampled_token_ids[0]) == 1
    # TKV shouldn't reflect any prefill padding, it's just the prompt
    assert output.tkv == prompt_len


@pytest.mark.cpu
@pytest.mark.chunked_prefill
@default_test_params
def test_multi_chunk_unpadded_prefill(
    model: ModelInfo,
    max_model_len: int,
    max_num_seqs: int,
    max_num_batched_tokens: int,
    monkeypatch: pytest.MonkeyPatch,
):
    """A request that's longer than a chunk can be split into multiple chunks
    with no padding required when the prompt is within one block of the end of
    a chunk"""
    runner: SpyreModelRunner = get_cpu_model_runner(
        model=model,
        max_model_len=max_model_len,
        max_num_seqs=max_num_seqs,
        max_num_batched_tokens=max_num_batched_tokens,
        monkeypatch=monkeypatch,
    )

    req_id = "1"
    # Prompt is within one block of two full chunks
    prompt_len = 2 * max_num_batched_tokens - 20
    new_req_data = make_new_request_data(req_id, prompt_len)

    # Scheduler will give first 128 token chunk
    scheduler_output = make_scheduler_output(
        scheduled_new_reqs=[new_req_data], num_scheduled_tokens={req_id: max_num_batched_tokens}
    )
    output = runner.execute_model(scheduler_output)

    # no output tokens
    assert len(output.sampled_token_ids) == 0
    # No left padding block so all 128 tokens are computed
    # TODO: uncomment when supported
    # assert output.computed_tokens[req_id] == max_num_batched_tokens
    computed_tokens = {req_id: max_num_batched_tokens}

    # Scheduler schedules remaining 200 - 128 = 72 tokens
    cached_req_data = make_cached_request_data(computed_tokens)
    scheduler_output = make_scheduler_output(
        scheduled_new_reqs=[],
        scheduled_cached_reqs=cached_req_data,
        num_scheduled_tokens={req_id: prompt_len - max_num_batched_tokens},
    )
    output = runner.execute_model(scheduler_output)

    # Should be one output token now
    assert len(output.sampled_token_ids[0]) == 1
    assert output.tkv == prompt_len


@pytest.mark.cpu
@pytest.mark.chunked_prefill
@default_test_params
def test_decode_padding_to_same_block(
    model: ModelInfo,
    max_model_len: int,
    max_num_seqs: int,
    max_num_batched_tokens: int,
    monkeypatch: pytest.MonkeyPatch,
):
    """Test that decode batches will use full blocks of left-padding to align
    themselves into the same last block of tokens in the sequence"""
    runner: SpyreModelRunner = get_cpu_model_runner(
        model=model,
        max_model_len=max_model_len,
        max_num_seqs=max_num_seqs,
        max_num_batched_tokens=max_num_batched_tokens,
        monkeypatch=monkeypatch,
    )

    short_req_id = "short"
    long_req_id = "long"
    short_prompt_len = 62
    long_prompt_len = 63
    steps = 0
    computed_tokens = lambda: {
        short_req_id: short_prompt_len + steps,
        long_req_id: long_prompt_len + steps,
    }

    # Prefill both in one pass each
    new_req_data = make_new_request_data(short_req_id, short_prompt_len)
    scheduler_output = make_scheduler_output(
        scheduled_new_reqs=[new_req_data], num_scheduled_tokens={short_req_id: short_prompt_len}
    )
    output = runner.execute_model(scheduler_output)

    new_req_data = make_new_request_data(long_req_id, long_prompt_len)
    scheduler_output = make_scheduler_output(
        scheduled_new_reqs=[new_req_data], num_scheduled_tokens={long_req_id: long_prompt_len}
    )
    output = runner.execute_model(scheduler_output)
    computed_tokens_dict = computed_tokens()

    # Save the number of free blocks to compare once we allocate a new one
    initial_free_blocks = runner.block_pool.get_num_free_blocks()

    # Both requests are still in the first block
    # Scheduler schedules 1 token each
    steps += 1
    cached_req_data = make_cached_request_data(computed_tokens_dict)
    scheduler_output = make_scheduler_output(
        scheduled_new_reqs=[],
        scheduled_cached_reqs=cached_req_data,
        num_scheduled_tokens={short_req_id: 1, long_req_id: 1},
    )
    output = runner.execute_model(scheduler_output)
    # TKV is the length of the long request since both are still in first block
    assert output.tkv == long_prompt_len + steps
    computed_tokens_dict = computed_tokens()

    # Now the long request is at the first block boundary (tkv = 64)
    # We'll have to pad the short request to be within the same block by adding
    # a full block of left padding to it
    steps += 1
    cached_req_data = make_cached_request_data(computed_tokens_dict)
    scheduler_output = make_scheduler_output(
        scheduled_new_reqs=[],
        scheduled_cached_reqs=cached_req_data,
        num_scheduled_tokens={short_req_id: 1, long_req_id: 1},
    )
    output = runner.execute_model(scheduler_output)
    # 🌶️🌶️🌶️ short prompt gets padded, it's now the longest sequence
    assert output.tkv == short_prompt_len + steps + 64
    # We should have allocated only one new block for the long prompt entering
    assert runner.block_pool.get_num_free_blocks() == initial_free_blocks - 1
    computed_tokens_dict = computed_tokens()

    # The shorter request is now at the second block boundary (tkv = 128), so we
    # can remove the extra block of padding (tkv = 64) since it will now be in
    # the second block along with the longer request
    steps += 1
    cached_req_data = make_cached_request_data(computed_tokens_dict)
    scheduler_output = make_scheduler_output(
        scheduled_new_reqs=[],
        scheduled_cached_reqs=cached_req_data,
        num_scheduled_tokens={short_req_id: 1, long_req_id: 1},
    )
    output = runner.execute_model(scheduler_output)
    # 🌶️🌶️🌶️ short prompt padding removed again, tkv is back to long + steps
    assert output.tkv == long_prompt_len + steps
    # One more real block was allocated for the short request
    assert runner.block_pool.get_num_free_blocks() == initial_free_blocks - 2



""" 
This example shows how to use Spyre with vLLM for running online inference,
using granite vision. Note that currently, multimodal is *not* supported for
static baching.

First, start the server with the following command:

VLLM_SPYRE_USE_CB=1 \
VLLM_SPYRE_DYNAMO_BACKEND=<your backend, e.g., sendnn/eager> \
vllm serve 'ibm-granite/granite-vision-3.3-2b' \
    --max-model-len=16384 \
    --max-num-seqs=2

NOTE: in the max feature case, a single image for granite vision can take
around 5k tokens, so keep this in mind when setting the max model length.
Also, although you should configure this to run as CB (as static batch is
not supported), this script does *not* submit multiple requests as a batch.
This is because multimodal inputs are only supported for chat completions,
not completions, and the chat completions endpoint does not support batched
inputs.
"""

import argparse
import time

from openai import OpenAI

parser = argparse.ArgumentParser(
    description="Script to submit an inference request to vllm server."
)

parser.add_argument(
    "--max_tokens",
    type=int,
    default=8,
    help="Maximum tokens.",
)
parser.add_argument(
    "--num_prompts",
    type=int,
    default=4,
)
parser.add_argument(
    "--stream",
    action=argparse.BooleanOptionalAction,
    help="Whether to stream the response.",
)

args = parser.parse_args()

# Modify OpenAI's API key and API base to use vLLM's API server.
openai_api_key = "EMPTY"
openai_api_base = "http://localhost:8000/v1"

client = OpenAI(
    # defaults to os.environ.get("OPENAI_API_KEY")
    api_key=openai_api_key,
    base_url=openai_api_base,
)


def get_vllm_prompts(num_prompts):
    """Get the vLLM prompts to be processed."""
    img_urls = [
        "https://vllm-public-assets.s3.us-west-2.amazonaws.com/vision_model_images/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg",  # noqa: E501
        "https://vllm-public-assets.s3.us-west-2.amazonaws.com/multimodal_asset/duck.jpg",  # noqa: E501
    ]

    instructions = [
        "describe this image.",
        "what is shown in this image?",
        "are there any animals in this image?",
    ]

    prompts = []
    for img_url in img_urls:
        for instr in instructions:
            prompts.append(
                {
                    "role": "user",
                    "content": [
                        {"type": "text", "text": instr},
                        {
                            "type": "image_url",
                            "image_url": {"url": img_url},
                        },
                    ],
                }
            )

    prompts = prompts * (num_prompts // len(prompts) + 1)
    return prompts[:num_prompts]


models = client.models.list()
model = models.data[0].id

prompts = get_vllm_prompts(args.num_prompts)

for prompt in prompts:
    stream = args.stream

    print(f"Prompt: {prompt}")
    start_t = time.time()

    chat_completion = client.chat.completions.create(
        messages=[prompt],
        model=model,
        max_completion_tokens=args.max_tokens,
        stream=stream,
    )

    end_t = time.time()
    print("Results:")
    if stream:
        for c in chat_completion:
            print(c.choices[0].delta.content, end="")
    else:
        print(chat_completion.choices[0].message.content)

    total_t = end_t - start_t
    print(f"\nDuration: {total_t}s")
